# keep line below to alias "main" module to icclim
from . import main as icclim

# keep line below to expose "main" content in icclim package namespace
from .main import index, indice

__version__ = "5.0.0rc2"
